#TODO: Start here... Needs to match interfaces

from .Interfaces import *
from enum import Enum
from pathlib import Path
import clr
import csv
clr.AddReference('System.Data')
import System
from System.Data import DataTable
from System.Data import DataRow
from System.Data import DataColumn
from System.Data import ParameterDirection

class TextFileFormat(Enum):
	CommaWithHeader = 1
	CommaWithoutHeader = 2
	TabWithHeader = 3
	TabWithoutHeader = 4
	PipeWithHeader = 5
	PipeWithoutHeader = 6
	Prettify = 7

class TextQueries(IQueries):
	FilePath:Path = None
	Format:TextFileFormat = TextFileFormat.CommaWithHeader

	def __init__(self, filePath:Path, format:TextFileFormat = TextFileFormat.CommaWithHeader):
		self.FilePath = filePath
		self.Format = format

	def Execute(self, query:str, parameters = None):
		raise NotImplementedError()

	def ExecuteWithScalar(self, query:str, parameters = None):
		raise NotImplementedError()

	def TruncateTable(self):
		header:str = None
		contents:str = self.FilePath.read_text()
		match self.Format:
			case TextFileFormat.CommaWithHeader:
				header = contents.split("\n")[0]
			case TextFileFormat.CommaWithoutHeader:
				header = ""

			case TextFileFormat.TabWithHeader:
				header = contents.split("\n")[0]
			case TextFileFormat.TabWithoutHeader:
				header = ""

			case TextFileFormat.PipeWithHeader:
				header = contents.split("\n")[0]
			case TextFileFormat.PipeWithoutHeader:
				header = ""
		self.FilePath.write_text(header)

	def GetRowCount(self) -> int:
		returnValue:int = None
		if (self.Format != TextFileFormat.Prettify):
			contents:str = self.FilePath.read_text()
			returnValue = len(contents.split("\n"))
			returnValue -= 1
			match self.Format:
				case TextFileFormat.CommaWithHeader:
					returnValue -= 1
				case TextFileFormat.TabWithHeader:
					returnValue -= 1
				case TextFileFormat.PipeWithHeader:
					returnValue -= 1
		return returnValue

class TextBulkData:
	FilePath:Path = None
	Format:TextFileFormat = TextFileFormat.CommaWithHeader

	def __init__(self, filePath:Path, format:TextFileFormat = TextFileFormat.CommaWithHeader):
		self.FilePath = filePath
		self.Format = format

	def __GetDataTableDelimited__(self, rowDelimiter:str, fieldDelimiter:str, includeHeader:bool) -> DataTable:
		returnValue:DataTable = None
		with open(self.FilePath, newline='') as file:
			returnValue = DataTable()
			reader = csv.reader(file, delimiter=fieldDelimiter, lineterminator=rowDelimiter)
			rowNumber:int = 0
			for reader in reader:
				rowNumber += 1
				if (rowNumber == 1):
					columns:list = []
					if (includeHeader):
						for item in reader:
							columnName:str = item
							dupeIndex:int = 0
							while (columns.count(columnName) > 0):
								dupeIndex += 1
								columnName = f"{columnName}_{dupeIndex}"
							columns.append(columnName)
					else:
						columns = ["Field_{}".format(x, y) for (x, y) in enumerate(reader)]
					for columnName in columns:
						dataColumn:DataColumn = DataColumn(columnName)
						dataColumn.DataType = System.Type.GetType("System.String")
						dataColumn.AllowDBNull  = True
						returnValue.Columns.Add(dataColumn)
					if (not includeHeader):
						dataRow:DataRow = returnValue.NewRow()
						for index, item in enumerate(reader):
							dataRow[index] = item
						returnValue.Rows.Add(dataRow)
				else:
					dataRow:DataRow = returnValue.NewRow()
					for index, item in enumerate(reader):
						dataRow[index] = item
					returnValue.Rows.Add(dataRow)
		return returnValue

	def GetDataTable(self) -> DataTable:
		returnValue:DataTable | None = None
		match self.Format:
			case TextFileFormat.CommaWithHeader:
				returnValue = self.__GetDataTableDelimited__("\n", ",", True)
			case TextFileFormat.CommaWithoutHeader:
				returnValue = self.__GetDataTableDelimited__("\n", ",", False)

			case TextFileFormat.TabWithHeader:
				returnValue = self.__GetDataTableDelimited__("\n", "\t", True)
			case TextFileFormat.TabWithoutHeader:
				returnValue = self.__GetDataTableDelimited__("\n", "\t", False)

			case TextFileFormat.PipeWithHeader:
				returnValue = self.__GetDataTableDelimited__("\n", "|", True)
			case TextFileFormat.PipeWithoutHeader:
				returnValue = self.__GetDataTableDelimited__("\n", "|", False)
		return returnValue

	def __WriteDataTableDelimited__(self, dataTable:DataTable, rowDelimiter:str, fieldDelimiter:str, includeHeader:bool):
		returnValue:str = None
		exception = None
		try:
			lines:list = []
			line:str = None
			line = ""
			if (includeHeader):
				for dataColumn in dataTable.Columns:
					line += f"{dataColumn.ColumnName}{fieldDelimiter}"
				if (line.endswith(fieldDelimiter)):
					line = line[:-len(fieldDelimiter)]
				lines.append(line)
				line = ""
			for dataRow in dataTable.Rows:
				line = ""
				for dataColumn in dataTable.Columns:
					if (str(dataColumn.DataType) in [
						"System.Byte",
						"System.Int16",
						"System.Int32",
						"System.Int64",
						"System.Decimal"]):
						line += f"{str(dataRow[dataColumn.ColumnName])}{fieldDelimiter}"
					else:
						line += f"{str(dataRow[dataColumn.ColumnName])}{fieldDelimiter}"
				if (line.endswith(fieldDelimiter)):
					line = line[:-len(fieldDelimiter)]
				lines.append(line)
			returnValue = ""
			for ln in lines:
				returnValue += f"{ln}{rowDelimiter}"
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		elif returnValue is not None:
			return returnValue

	def __WriteDataTablePrettify__(self, dataTable:DataTable):
		returnValue:str = None
		exception = None
		try:
			totalColumnWidth:int = 0
			columnWidth:dict = {}
			lines:list = []
			line:str = None
			allLines:str = ""
			for dataRow in dataTable.Rows:
				for dataColumn in dataTable.Columns:
					if (dataColumn.ColumnName not in columnWidth):
						columnWidth[dataColumn.ColumnName] = len(dataColumn.ColumnName)
					if (columnWidth[dataColumn.ColumnName] < len(str(dataRow[dataColumn]))):
						columnWidth[dataColumn.ColumnName] = len(str(dataRow[dataColumn]))
			for columnName in columnWidth:
				totalColumnWidth += columnWidth[columnName]
			lines.append(" -" + ("-"*totalColumnWidth) + ("--"*len(columnWidth)) + "- ")
			line = ""
			for columnName in columnWidth:
				line += f" {columnName.ljust(columnWidth[columnName])} "
			lines.append(f"| {line} |")
			line = ""
			for columnName in columnWidth:
				headBreak:str = "-"*columnWidth[columnName]
				line += f" {headBreak} "
			lines.append(f"| {line} |")
			for dataRow in dataTable.Rows:
				line = ""
				for dataColumn in dataTable.Columns:
					if (str(dataColumn.DataType) in [
						"System.Byte",
						"System.Int16",
						"System.Int32",
						"System.Int64",
						"System.Decimal"]):
						line += f" {str(dataRow[dataColumn.ColumnName]).rjust(columnWidth[dataColumn.ColumnName])} "
					else:
						line += f" {str(dataRow[dataColumn.ColumnName]).ljust(columnWidth[dataColumn.ColumnName])} "
				lines.append(f"| {line} |")
			lines.append(" -" + ("-"*totalColumnWidth) + ("--"*len(columnWidth)) + "- ")
			returnValue = ""
			for ln in lines:
				returnValue += f"{ln}\n"
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		elif returnValue is not None:
			return returnValue

	def WriteDataTable(self, dataTable:DataTable):
		exception = None
		try:
			text:str = None
			match self.Format:
				case TextFileFormat.CommaWithHeader:
					text = self.__WriteDataTableDelimited__(dataTable, "\n", ",", True)
				case TextFileFormat.CommaWithoutHeader:
					text = self.__WriteDataTableDelimited__(dataTable, "\n", ",", False)
				case TextFileFormat.TabWithHeader:
					text = self.__WriteDataTableDelimited__(dataTable, "\n", "\t", True)
				case TextFileFormat.TabWithoutHeader:
					text = self.__WriteDataTableDelimited__(dataTable, "\n", "\t", False)
				case TextFileFormat.PipeWithHeader:
					text = self.__WriteDataTableDelimited__(dataTable, "\n", "|", True)
				case TextFileFormat.PipeWithoutHeader:
					text = self.__WriteDataTableDelimited__(dataTable, "\n", "|", False)
				case TextFileFormat.Prettify:
					text = self.__WriteDataTablePrettify__(dataTable)

			if (text is not None):
				with open(self.FilePath, "w") as file:
					file.write(text)
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception

class TextClearData:
	FilePath:Path = None
	Format:TextFileFormat = TextFileFormat.CommaWithHeader

	def __init__(self, filePath:Path, format:TextFileFormat = TextFileFormat.CommaWithHeader):
		self.FilePath = filePath
		self.Format = format
		self.Queries = TextQueries(filePath, format)
		self.BulkData = TextBulkData(filePath, format)

__all__ = ["TextFileFormat", "TextBulkData", "TextQueries", "TextClearData"]



from .Interfaces import *
import psycopg
from psycopg.rows import DictRow, dict_row
import datetime
import clr
import json
from pathlib import Path
clr.AddReference('System.Data')
import System
from System.Data import DataTable
from System.Data import DataRow
from System.Data import DataColumn
from System.Data import ParameterDirection

class PostgreSQLParameter:
	def __init__(self, name:str, dataType:str, direction:ParameterDirection, value, size:int = None):
		self.Name = name
		self.Direction = direction
		self.Value = value

class PostgreSQLStoredProcedures(IStoredProcedures):
	ConnectionString:str = None

	def __init__(self, connectionString:str):
		self.ConnectionString = connectionString

	def Execute(self, schema:str, name:str, parameters:dict = None) -> None:
		exception = None
		try:
			parameterText:str = None
			if (parameters is None):
				parameterText = ""
			else:
				for key in parameters.keys():
					if (parameterText is None):
						parameterText = f"%({key})s"
					else:
						parameterText += f", %({key})s"
			commandText:str = f"SELECT * FROM \"{schema}\".\"{name}\"({parameterText}) AS \"output\""
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText, parameters)
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception

	def ExecuteWithScalar(self, schema:str, name:str, parameters:dict = None) -> any:
		returnValue:any | None = None
		exception = None
		try:
			parameterText:str = None
			if (parameters is None):
				parameterText = ""
			else:
				for key in parameters.keys():
					if (parameterText is None):
						parameterText = f"%({key})s"
					else:
						parameterText += f", %({key})s"
			commandText:str = f"SELECT * FROM \"{schema}\".\"{name}\"({parameterText}) AS \"output\""
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText, parameters)
					for row in cur:
						returnValue = row[0]
						break
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		else:
			return returnValue

	def ExecuteJSON(self,
			schema:str,
			name:str,
			inputValue:str = None
		) -> str:
		returnValue:str | None = None
		exception = None
		try:
			commandText:str = f"SELECT FROM \"{schema}\".\"{name}\"(%(Input)s) AS \"output\""
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText, {"Input": inputValue})
					for row in cur:
						returnValue = row[0]
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		else:
			return returnValue

	def ExecuteJSONInputOnly(self,
				 schema:str,
				 name:str,
				 inputValue:str = None,
		) -> None:
		exception = None
		try:
			commandText:str = f"SELECT FROM \"{schema}\".\"{name}\"(%(Input)s) AS \"output\""
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText, {"Input": inputValue})
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception

	def ExecuteJSONOutputOnly(self,
				 schema:str,
				 name:str
		) -> str:
		returnValue:str | None = None
		exception = None
		try:
			commandText:str = f"SELECT FROM \"{schema}\".\"{name}\"() AS \"output\""
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText)
					for row in cur:
						returnValue = row[0]
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		else:
			return returnValue

	def ExecuteDataTable(self, schema:str, name:str, parameters:dict = None) -> DataTable:
		returnValue = None
		exception = None
		try:
			parameterText:str = None
			if (parameters is None):
				parameterText = ""
			else:
				for key in parameters.keys():
					keyValue:str = "NULL"
					if (parameters[key] is not None): keyValue = f"'{parameters[key]}'"
					keyValue = f"\"{key}\"=>{keyValue}"
					if (parameterText is None):
						parameterText = f"{keyValue}"
					else:
						parameterText += f", {keyValue}"
			commandText:str = f"SELECT * FROM \"{schema}\".\"{name}\"({parameterText}) AS \"output\""
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText, parameters)
					returnValue = DataTable()
					rowCount:int = 0
					for row in cur:
						rowCount += 1
						dataRow = returnValue.NewRow()
						for columnIndex, column in enumerate(cur.description):
							clrDataType:str = PostgreSQLClearData.__GetCLRDataType__(row[columnIndex])
							if (not returnValue.Columns.Contains(column.name)):
								dataColumn:DataColumn = DataColumn(column.name)
								dataColumn.DataType = System.Type.GetType(clrDataType)
								dataColumn.AllowDBNull  = True
								#returnValue.Columns.Add(column.name, System.Type.GetType(clrDataType))
								returnValue.Columns.Add(dataColumn)
							dataRow[columnIndex] = PostgreSQLClearData.__ConvertToCLRDataType__(row[columnIndex], clrDataType)
						returnValue.Rows.Add(dataRow)

		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		else:
			return returnValue

class PostgreSQLQueries(IQueries):
	ConnectionString:str = None

	def __init__(self, connectionString:str):
		self.ConnectionString = connectionString

	def Execute(self, query:str):
		exception = None
		try:
			commandText:str = query
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText)
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception

	def ExecuteWithScalar(self, query:str):
		returnValue = None
		exception = None
		try:
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(query)
					for row in cur:
						returnValue = row[0]
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		if returnValue is not None:
			return returnValue

	def TruncateTable(self, schema:str, table:str):
		exception = None
		try:
			commandText:str = f"TRUNCATE TABLE {schema}.{table}"
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText)
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception

	def GetRowCount(self, schema:str, table:str) -> int:
		returnValue:int = None
		exception = None
		try:
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(f"SELECT COUNT(*) AS \"Count\" FROM \"{schema}\".\"{table}\"")
					returnValue = int(cur.fetchone()[0])
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		if returnValue is not None:
			return returnValue

class PostgreSQLBulkData(IBulkData):
	ConnectionString:str = None

	def __init__(self, connectionString:str):
		self.ConnectionString = connectionString

	def GetDataTable(
		self,
		schema:str = None,
		tableOrView:str = None,
		query:str = None,
		options:dict = {"UseBinaryFormat": False, "CLRTypeOverride": {} }
	):
		returnValue = None
		exception = None
		commandText:str = None
		if (schema is not None and tableOrView is not None):
			commandText = f"SELECT * FROM \"{schema}\".\"{tableOrView}\""
		elif (query is not None):
			commandText = query
		if (commandText is None):
			raise ValueError("Either both Schema and Table/View must be provided, or Query must be provided.")

		useBinaryFormat:bool = False
		clrTypeOverride:dict = {}
		if ("BatchSize" in options):
			useBinaryFormat = bool(options["UseBinaryFormat"])
		if ("CLRTypeOverride" in options):
			if (isinstance(options["CLRTypeOverride"], dict)):
				clrTypeOverride = options["CLRTypeOverride"]
		try:
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText)
					returnValue = DataTable()
					rowCount:int = 0
					for row in cur:
						rowCount += 1
						dataRow = returnValue.NewRow()
						for columnIndex, column in enumerate(cur.description):
							clrDataType:str = None
							if (column.name not in clrTypeOverride):
								clrDataType:str = PostgreSQLClearData.__GetCLRDataType__(row[columnIndex])
							else:
								clrDataType = clrTypeOverride[column.name]
							if (not returnValue.Columns.Contains(column.name)):
								dataColumn:DataColumn = DataColumn(column.name)
								dataColumn.DataType = System.Type.GetType(clrDataType)
								dataColumn.AllowDBNull  = True
								#returnValue.Columns.Add(column.name, System.Type.GetType(clrDataType))
								returnValue.Columns.Add(dataColumn)
							dataRow[columnIndex] = PostgreSQLClearData.__ConvertToCLRDataType__(row[columnIndex], clrDataType)
						returnValue.Rows.Add(dataRow)
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		elif returnValue is not None:
			return returnValue

	def WriteDataTable(self, schema:str, table:str, dataTable:DataTable, options:dict = {"UseBinaryFormat": False}):
		exception = None
		binaryFormat:str = ""
		useBinaryFormat:bool = False
		if ("BatchSize" in options):
			useBinaryFormat = bool(options["UseBinaryFormat"])
		if (useBinaryFormat):
			binaryFormat = " (FORMAT BINARY)"
		try:
			columnList:str = None
			for column in dataTable.Columns:
				if (columnList is None):
					columnList = f"\"{column.ColumnName}\""
				else:
					columnList += f", \"{column.ColumnName}\""
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					returnValue = DataTable()
					with cur.copy(f"COPY \"{schema}\".\"{table}\" ({columnList}) FROM STDOUT{binaryFormat}") as copy:
						for row in dataTable.Rows:
							rowData = []
							for item in row.ItemArray:
								if (isinstance(item, System.DBNull)):
									rowData.append(None)
								elif (isinstance(item, System.DateTime)):
									rowData.append(datetime.datetime.fromisoformat(item.ToString("yyyy-MM-ddTHH:mm:ss.fffffff")))
								else:
									rowData.append(item)
							copy.write_row(rowData)
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception

class PostgreSQLClearData:
	ConnectionString:str = None

	def __init__(self, connectionString:str):
		self.ConnectionString = connectionString
		self.StoredProcedures = PostgreSQLStoredProcedures(connectionString)
		self.Queries = PostgreSQLQueries(connectionString)
		self.BulkData = PostgreSQLBulkData(connectionString)

	def __GetTestConnectionQuery__(self) -> str:
		return """
			SELECT ROW_TO_JSON(\"Sections\")
				FROM
				(
					SELECT
						(
							SELECT ROW_TO_JSON(\"Client\")
								FROM
								(
									SELECT
										CURRENT_SETTING('application_name') AS \"ApplicationName\",
										inet_client_addr() AS \"HostAddress\",
										inet_client_port() AS \"HostPort\",
										CURRENT_ROLE AS \"Role\",
										CURRENT_USER AS \"CurrentUser\",
										USER AS \"User\"
								) AS \"Client\"
						) AS \"Client\",
						(
							SELECT ROW_TO_JSON(\"Server\")
								FROM
								(
									SELECT
										inet_server_addr() AS \"HostAddress\",
										inet_server_port() AS \"HostPort\",
										CURRENT_SETTING('SERVER_VERSION') AS \"Version\"
								) AS \"Server\"
						) AS \"Server\",
						(
							SELECT ROW_TO_JSON(\"Database\")
								FROM
								(
									SELECT
										current_database()  AS \"Name\"
								) AS \"Database\"
						) AS \"Database\"
				) AS \"Sections\"
		"""

	def TestConnection(self) -> dict:
		returnValue:dict = None
		commandText:str = self.__GetTestConnectionQuery__()
		with open(Path(__file__).parent.joinpath("PostgreSQLFunctions.TestConnection.sql"), "r") as file:
			commandText = file.read()
		exception = None
		try:
			with psycopg.connect(self.ConnectionString) as conn:
				with conn.cursor() as cur:
					cur.execute(commandText)
					result = cur.fetchone()[0]
					if (result):
						returnValue = json.loads(result)
		except Exception as e:
			exception = e
		if exception is not None:
			raise exception
		return returnValue

	@staticmethod
	def __GetDataColumn__(name:str, clrDataType:str):
		dataColumn:DataColumn = DataColumn(name)
		dataColumn = System.Type.GetType(clrDataType)
		return dataColumn

	@staticmethod
	def __GetCLRDataType__(obj):
		if (isinstance(obj, bool)):
			return "System.Boolean"
		elif (isinstance(obj, datetime.datetime)):
			return "System.DateTime"
		elif (isinstance(obj, int)):
			if (obj > 2147483647):
				return "System.Int64"
			else:
				return "System.Int32"
		else:
			return "System.String"

	@staticmethod
	def __ConvertToCLRDataType__(obj, exceptedType:str = None):
		returnValue = None
		if (obj is None):
			returnValue = System.DBNull.Value
		else:
			if (exceptedType is None):
				exceptedType = BulkData.__GetCLRDataType__(obj)
			match exceptedType:
				case "System.Boolean":
					returnValue = System.Convert.ToBoolean(obj)
				case "System.DateTime":
					returnValue = System.Convert.ToDateTime(obj.isoformat())
				case "System.Int32":
					returnValue = System.Convert.ToInt32(obj)
				case "System.Int64":
					returnValue = System.Convert.ToInt64(obj)
				case _:
					returnValue = System.Convert.ToString(obj)
		return returnValue

__all__ = ["PostgreSQLParameter", "PostgreSQLStoredProcedures", "PostgreSQLQueries", "PostgreSQLBulkData", "PostgreSQLClearData"]



from .Interfaces import *
import clr
import json
from pathlib import Path
clr.AddReference('System.Data')
from System import DateTime
from System.Data import *
from System.Data import CommandType
from System.Data import SqlClient
from System.Data import SqlDbType
from System.Data import ParameterDirection
from System.Data import DataTable

class SQLServerParameter:
	def __init__(self, name:str, sqlDataType:SqlDbType, direction:ParameterDirection, value = None, size:int = None):
		self.Name = name
		self.SqlDbType = sqlDataType
		self.Direction = direction
		self.Value = value
		self.Size = size

	@staticmethod
	def CreateBitInput(name:str, value:bool):
		return SQLServerParameter(name, SqlDbType.Bit, ParameterDirection.Input, value, None)

	@staticmethod
	def CreateBitOutput(name:str):
		return SQLServerParameter(name, SqlDbType.Bit, ParameterDirection.Output, None, None)

	@staticmethod
	def CreateIntInput(name:str, value:int):
		return SQLServerParameter(name, SqlDbType.Int, ParameterDirection.Input, value, None)

	@staticmethod
	def CreateIntOutput(name:str):
		return SQLServerParameter(name, SqlDbType.Int, ParameterDirection.Output, None, None)

	@staticmethod
	def CreateBigIntInput(name:str, value:int):
		return SQLServerParameter(name, SqlDbType.BigInt, ParameterDirection.Input, value, None)

	@staticmethod
	def CreateBigIntOutput(name:str):
		return SQLServerParameter(name, SqlDbType.BigInt, ParameterDirection.Output, None, None)

	@staticmethod
	def CreateNVarCharInput(name:str, size:int, value:str):
		return SQLServerParameter(name, SqlDbType.NVarChar, ParameterDirection.Input, value, size)

	@staticmethod
	def CreateNVarCharOutput(name:str, size:int):
		return SQLServerParameter(name, SqlDbType.NVarChar, ParameterDirection.Output, None, size)

	@staticmethod
	def CreateVarCharInput(name:str, size:int, value:str):
		return SQLServerParameter(name, SqlDbType.VarChar, ParameterDirection.Input, value, size)

	@staticmethod
	def CreateVarCharOutput(name:str, size:int):
		return SQLServerParameter(name, SqlDbType.VarChar, ParameterDirection.Output, None, size)

	@staticmethod
	def CreateSysNameInput(name:str, value:str):
		return SQLServerParameter(name, SqlDbType.NVarChar, ParameterDirection.Input, value, 128)

	@staticmethod
	def CreateSysNameOutput(name:str):
		return SQLServerParameter(name, SqlDbType.NVarChar, ParameterDirection.Output, None, 128)

	@staticmethod
	def CreateDateTime2Input(name:str, size:int, value:DateTime):
		if size not in range(0, 7):
			raise ValueError('Size must be an integer between 0 and 7.')
		return SQLServerParameter(name, SqlDbType.DateTime2, ParameterDirection.Input, value, size)

	@staticmethod
	def CreateDateTime2Output(sqlCommand:SqlClient.SqlCommand, name:str, size:int):
		if size not in range(0, 7):
			raise ValueError('Size must be an integer between 0 and 7.')
		return SQLServerParameter(name, SqlDbType.DateTime2, ParameterDirection.Output, None, size)

class _ParametersHelpers:
	@staticmethod
	def GetParameter(sqlCommand:SqlClient.SqlCommand, name:str):
		returnValue = [ p for p in sqlCommand.Parameters if p.ParameterName == name ][0]
		return returnValue

	@staticmethod
	def AddParameters(sqlCommand:SqlClient.SqlCommand, parameters:list[any]):
		for parameter in parameters:
			if (type(parameter).__name__ != "SQLServerParameter"):
				raise ValueError(f"paramteres must be a list of SQLServerParameter objects. {type(parameter).__name__} is not valid.")
			match parameter.SqlDbType:
				case SqlDbType.Bit:
					if parameter.Direction == ParameterDirection.Input:
						_ParametersHelpers.AddBitInput(sqlCommand, parameter.Name, parameter.Value)
					elif parameter.Direction == ParameterDirection.Output:
						_ParametersHelpers.AddBitOutput(sqlCommand, parameter.Name)
				case SqlDbType.Int:
					if parameter.Direction == ParameterDirection.Input:
						_ParametersHelpers.AddIntInput(sqlCommand, parameter.Name, parameter.Value)
					elif parameter.Direction == ParameterDirection.Output:
						_ParametersHelpers.AddIntOutput(sqlCommand, parameter.Name)
				case SqlDbType.BigInt:
					if parameter.Direction == ParameterDirection.Input:
						_ParametersHelpers.AddBigIntInput(sqlCommand, parameter.Name, parameter.Value)
					elif parameter.Direction == ParameterDirection.Output:
						_ParametersHelpers.AddBigIntOutput(sqlCommand, parameter.Name)
				case SqlDbType.NVarChar:
					if parameter.Direction == ParameterDirection.Input:
						_ParametersHelpers.AddNVarCharInput(sqlCommand, parameter.Name, parameter.Size, parameter.Value)
					elif parameter.Direction == ParameterDirection.Output:
						_ParametersHelpers.AddNVarCharOutput(sqlCommand, parameter.Name, parameter.Size)
				case SqlDbType.VarChar:
					if parameter.Direction == ParameterDirection.Input:
						_ParametersHelpers.AddVarCharInput(sqlCommand, parameter.Name, parameter.Size, parameter.Value)
					elif parameter.Direction == ParameterDirection.Output:
						_ParametersHelpers.AddVarCharOutput(sqlCommand, parameter.Name, parameter.Size)
				case SqlDbType.DateTime2:
					if parameter.Direction == ParameterDirection.Input:
						_ParametersHelpers.AddDateTime2Input(sqlCommand, parameter.Name, parameter.Size, parameter.Value)
					elif parameter.Direction == ParameterDirection.Output:
						_ParametersHelpers.AddDateTime2Output(sqlCommand, parameter.Name, parameter.Size)
				case _:
					raise TypeError(f"Unhandled SqlDbType ({parameter.SqlDbType})")

	@staticmethod
	def AddBitInput(sqlCommand:SqlClient.SqlCommand, name:str, value:bool):
		sqlParameter = sqlCommand.CreateParameter()
		sqlParameter.ParameterName = name
		sqlParameter.SqlDbType = SqlDbType.Bit
		sqlParameter.Direction = ParameterDirection.Input
		sqlParameter.SqlValue = value
		sqlCommand.Parameters.Add(sqlParameter)

	@staticmethod
	def AddBitOutput(sqlCommand:SqlClient.SqlCommand, name:str):
		sqlParameter = sqlCommand.CreateParameter()
		sqlParameter.ParameterName = name
		sqlParameter.SqlDbType = SqlDbType.Bit
		sqlParameter.Direction = ParameterDirection.Output
		sqlCommand.Parameters.Add(sqlParameter)

	@staticmethod
	def AddIntInput(sqlCommand:SqlClient.SqlCommand, name:str, value:int):
		sqlParameter = sqlCommand.CreateParameter()
		sqlParameter.ParameterName = name
		sqlParameter.SqlDbType = SqlDbType.Int
		sqlParameter.Direction = ParameterDirection.Input
		sqlParameter.SqlValue = value
		sqlCommand.Parameters.Add(sqlParameter)

	@staticmethod
	def AddIntOutput(sqlCommand:SqlClient.SqlCommand, name:str):
		sqlParameter = sqlCommand.CreateParameter()
		sqlParameter.ParameterName = name
		sqlParameter.SqlDbType = SqlDbType.Int
		sqlParameter.Direction = ParameterDirection.Output
		sqlCommand.Parameters.Add(sqlParameter)

	@staticmethod
	def AddBigIntInput(sqlCommand:SqlClient.SqlCommand, name:str, value:int):
		sqlParameter = sqlCommand.CreateParameter()
		sqlParameter.ParameterName = name
		sqlParameter.SqlDbType = SqlDbType.BigInt
		sqlParameter.Direction = ParameterDirection.Input
		sqlParameter.SqlValue = value
		sqlCommand.Parameters.Add(sqlParameter)

	@staticmethod
	def AddBigIntOutput(sqlCommand:SqlClient.SqlCommand, name:str):
		sqlParameter = sqlCommand.CreateParameter()
		sqlParameter.ParameterName = name
		sqlParameter.SqlDbType = SqlDbType.BigInt
		sqlParameter.Direction = ParameterDirection.Output
		sqlCommand.Parameters.Add(sqlParameter)

	@staticmethod
	def AddNVarCharInput(sqlCommand:SqlClient.SqlCommand, name:str, size:int, value:str):
		sqlParameter = sqlCommand.CreateParameter()
		sqlParameter.ParameterName = name
		sqlParameter.SqlDbType = SqlDbType.NVarChar
		sqlParameter.Direction = ParameterDirection.Input
		sqlParameter.Size = size
		sqlParameter.SqlValue = value
		sqlCommand.Parameters.Add(sqlParameter)

	@staticmethod
	def AddNVarCharOutput(sqlCommand:SqlClient.SqlCommand, name:str, size:int):
		sqlParameter = sqlCommand.CreateParameter()
		sqlParameter.ParameterName = name
		sqlParameter.SqlDbType = SqlDbType.NVarChar
		sqlParameter.Direction = ParameterDirection.Output
		sqlParameter.Size = size
		sqlCommand.Parameters.Add(sqlParameter)

	@staticmethod
	def AddVarCharInput(sqlCommand:SqlClient.SqlCommand, name:str, size:int, value:str):
		sqlParameter = sqlCommand.CreateParameter()
		sqlParameter.ParameterName = name
		sqlParameter.SqlDbType = SqlDbType.VarChar
		sqlParameter.Direction = ParameterDirection.Input
		sqlParameter.Size = size
		sqlParameter.SqlValue = value
		sqlCommand.Parameters.Add(sqlParameter)

	@staticmethod
	def AddVarCharOutput(sqlCommand:SqlClient.SqlCommand, name:str, size:int):
		sqlParameter = sqlCommand.CreateParameter()
		sqlParameter.ParameterName = name
		sqlParameter.SqlDbType = SqlDbType.VarChar
		sqlParameter.Direction = ParameterDirection.Output
		sqlParameter.Size = size
		sqlCommand.Parameters.Add(sqlParameter)

	@staticmethod
	def AddSysNameInput(sqlCommand:SqlClient.SqlCommand, name:str, value:str):
		sqlParameter = sqlCommand.CreateParameter()
		sqlParameter.ParameterName = name
		sqlParameter.SqlDbType = SqlDbType.NVarChar
		sqlParameter.Direction = ParameterDirection.Input
		sqlParameter.Size = 128
		sqlParameter.SqlValue = value
		sqlCommand.Parameters.Add(sqlParameter)

	@staticmethod
	def AddSysNameOutput(sqlCommand:SqlClient.SqlCommand, name:str):
		sqlParameter = sqlCommand.CreateParameter()
		sqlParameter.ParameterName = name
		sqlParameter.SqlDbType = SqlDbType.NVarChar
		sqlParameter.Direction = ParameterDirection.Output
		sqlParameter.Size = 128
		sqlCommand.Parameters.Add(sqlParameter)

	@staticmethod
	def AddDateTime2Input(sqlCommand:SqlClient.SqlCommand, name:str, size:int, value:DateTime):
		if size not in range(0, 7):
			raise ValueError('Size must be an integer between 0 and 7.')
		sqlParameter = sqlCommand.CreateParameter()
		sqlParameter.ParameterName = name
		sqlParameter.SqlDbType = SqlDbType.DateTime2
		sqlParameter.Direction = ParameterDirection.Input
		sqlParameter.Size = size
		sqlParameter.SqlValue = value
		sqlCommand.Parameters.Add(sqlParameter)

	@staticmethod
	def AddDateTime2Output(sqlCommand:SqlClient.SqlCommand, name:str, size:int):
		if size not in range(0, 7):
			raise ValueError('Size must be an integer between 0 and 7.')
		sqlParameter = sqlCommand.CreateParameter()
		sqlParameter.ParameterName = name
		sqlParameter.SqlDbType = SqlDbType.DateTime2
		sqlParameter.Direction = ParameterDirection.Output
		sqlParameter.Size = size
		sqlCommand.Parameters.Add(sqlParameter)

class SQLServerStoredProcedures(IStoredProcedures):
	ConnectionString:str = None

	def __init__(self, connectionString:str):
		self.ConnectionString = connectionString

	def Execute(self, schema:str, name:str, parameters:list[any] = None):
		exception = None
		sqlConnection = None
		sqlCommand = None
		try:
			sqlConnection = SqlClient.SqlConnection(self.ConnectionString)
			sqlConnection.Open()
			sqlCommand = SqlClient.SqlCommand(f"[{schema}].[{name}]", sqlConnection)
			sqlCommand.CommandType = CommandType.StoredProcedure
			sqlCommand.CommandTimeout = 0
			if (parameters is not None):
				_ParametersHelpers.AddParameters(sqlCommand, parameters)
			sqlCommand.ExecuteNonQuery()
		except Exception as e:
			exception = e
		finally:
			if sqlCommand:
				sqlCommand.Dispose()
			if sqlConnection:
				sqlConnection.Close()
				sqlConnection.Dispose()
		if exception is not None:
			raise exception

	def ExecuteWithScalar(self, schema:str, name:str, parameters = None) -> any:
		returnValue = None
		exception = None
		sqlConnection = None
		sqlCommand = None
		try:
			sqlConnection = SqlClient.SqlConnection(self.ConnectionString)
			sqlConnection.Open()
			sqlCommand = SqlClient.SqlCommand(f"[{schema}].[{name}]", sqlConnection)
			sqlCommand.CommandType = CommandType.StoredProcedure
			sqlCommand.CommandTimeout = 0
			if (parameters is not None):
				_ParametersHelpers.AddParameters(sqlCommand, parameters)
			returnValue = sqlCommand.ExecuteScalar()
		except Exception as e:
			exception = e
		finally:
			if sqlCommand:
				sqlCommand.Dispose()
			if sqlConnection:
				sqlConnection.Close()
				sqlConnection.Dispose()
		if exception is not None:
			raise exception
		if returnValue is not None:
			return returnValue

	def ExecuteJSONNamed(self,
				 schema:str,
				 name:str,
				 inputParameterName:str = "Input",
				 inputValue:str = None,
				 outputParameterName:str = "Output"
		) -> str:
		returnValue:str = None
		exception = None
		sqlConnection = None
		sqlCommand = None
		try:
			sqlConnection = SqlClient.SqlConnection(self.ConnectionString)
			sqlConnection.Open()
			sqlCommand = SqlClient.SqlCommand(f"[{schema}].[{name}]", sqlConnection)
			sqlCommand.CommandType = CommandType.StoredProcedure
			sqlCommand.CommandTimeout = 0
			if (inputParameterName is not None):
				_ParametersHelpers.AddNVarCharInput(sqlCommand, inputParameterName, -1, inputValue)
			if (outputParameterName is not None):
				_ParametersHelpers.AddNVarCharOutput(sqlCommand, outputParameterName, -1)
			sqlCommand.ExecuteNonQuery()
			if (outputParameterName is not None):
				outputParameter = _ParametersHelpers.GetParameter(sqlCommand, outputParameterName)
				if (outputParameter):
					if (outputParameter.Value):
						if (type(outputParameter.Value) == str):
							returnValue = str(outputParameter.Value)
		except Exception as e:
			exception = e
		finally:
			if sqlCommand:
				sqlCommand.Dispose()
			if sqlConnection:
				sqlConnection.Close()
				sqlConnection.Dispose()
		if exception is not None:
			raise exception
		if returnValue is not None:
			return returnValue

	def ExecuteJSON(self,
			schema:str,
			name:str,
			inputValue:str = None
		) -> str:
		returnValue:str = None
		exception = None
		sqlConnection = None
		sqlCommand = None
		try:
			sqlConnection = SqlClient.SqlConnection(self.ConnectionString)
			sqlConnection.Open()
			sqlCommand = SqlClient.SqlCommand(f"[{schema}].[{name}]", sqlConnection)
			sqlCommand.CommandType = CommandType.StoredProcedure
			sqlCommand.CommandTimeout = 0
			_ParametersHelpers.AddNVarCharInput(sqlCommand, "Input", -1, inputValue)
			_ParametersHelpers.AddNVarCharOutput(sqlCommand, "Output", -1)
			sqlCommand.ExecuteNonQuery()
			outputParameter = _ParametersHelpers.GetParameter(sqlCommand, "Output")
			if (outputParameter):
				if (outputParameter.Value):
					if (type(outputParameter.Value) == str):
						returnValue = str(outputParameter.Value)
		except Exception as e:
			exception = e
		finally:
			if sqlCommand:
				sqlCommand.Dispose()
			if sqlConnection:
				sqlConnection.Close()
				sqlConnection.Dispose()
		if exception is not None:
			raise exception
		if returnValue is not None:
			return returnValue

	def ExecuteJSONInputOnly(self,
				 schema:str,
				 name:str,
				 inputValue:str = None,
		):
		exception = None
		sqlConnection = None
		sqlCommand = None
		try:
			sqlConnection = SqlClient.SqlConnection(self.ConnectionString)
			sqlConnection.Open()
			sqlCommand = SqlClient.SqlCommand(f"[{schema}].[{name}]", sqlConnection)
			sqlCommand.CommandType = CommandType.StoredProcedure
			sqlCommand.CommandTimeout = 0
			_ParametersHelpers.AddNVarCharInput(sqlCommand, "Input", -1, inputValue)
			sqlCommand.ExecuteNonQuery()
		except Exception as e:
			exception = e
		finally:
			if sqlCommand:
				sqlCommand.Dispose()
			if sqlConnection:
				sqlConnection.Close()
				sqlConnection.Dispose()
		if exception is not None:
			raise exception

	def ExecuteJSONOutputOnly(self,
				 schema:str,
				 name:str
		) -> str:
		returnValue:str = None
		exception = None
		sqlConnection = None
		sqlCommand = None
		try:
			sqlConnection = SqlClient.SqlConnection(self.ConnectionString)
			sqlConnection.Open()
			sqlCommand = SqlClient.SqlCommand(f"[{schema}].[{name}]", sqlConnection)
			sqlCommand.CommandType = CommandType.StoredProcedure
			sqlCommand.CommandTimeout = 0
			_ParametersHelpers.AddNVarCharOutput(sqlCommand, "Output", -1)
			sqlCommand.ExecuteNonQuery()
			outputParameter = _ParametersHelpers.GetParameter(sqlCommand, "Output")
			if (outputParameter):
				if (outputParameter.Value):
					if (type(outputParameter.Value) == str):
						returnValue = str(outputParameter.Value)
		except Exception as e:
			exception = e
		finally:
			if sqlCommand:
				sqlCommand.Dispose()
			if sqlConnection:
				sqlConnection.Close()
				sqlConnection.Dispose()
		if exception is not None:
			raise exception
		if returnValue is not None:
			return returnValue

	def ExecuteDataTable(self, schema:str, name:str, parameters = None) -> DataTable:
		returnValue = DataTable()
		exception = None
		sqlConnection = None
		sqlCommand = None
		try:
			sqlConnection = SqlClient.SqlConnection(self.ConnectionString)
			sqlConnection.Open()
			sqlCommand = SqlClient.SqlCommand(f"[{schema}].[{name}]", sqlConnection)
			sqlCommand.CommandType = CommandType.StoredProcedure
			sqlCommand.CommandTimeout = 0
			if (parameters is not None):
				_ParametersHelpers.AddParameters(sqlCommand, parameters)
			returnValue.Load(sqlCommand.ExecuteReader())
		except Exception as e:
			exception = e
		finally:
			if sqlCommand:
				sqlCommand.Dispose()
			if sqlConnection:
				sqlConnection.Close()
				sqlConnection.Dispose()
		if exception is not None:
			raise exception
		return returnValue

class SQLServerQueries(IQueries):
	ConnectionString:str = None

	def __init__(self, connectionString:str):
		self.ConnectionString = connectionString

	def Execute(self, query:str, parameters:list[any] = None):
		exception = None
		sqlConnection = None
		sqlCommand = None
		try:
			sqlConnection = SqlClient.SqlConnection(self.ConnectionString)
			sqlConnection.Open()
			sqlCommand = SqlClient.SqlCommand(query, sqlConnection)
			sqlCommand.CommandType = CommandType.Text
			sqlCommand.CommandTimeout = 0
			if (parameters is not None):
				_ParametersHelpers.AddParameters(sqlCommand, parameters)
			sqlCommand.ExecuteNonQuery()
		except Exception as e:
			exception = e
		finally:
			if sqlCommand:
				sqlCommand.Dispose()
			if sqlConnection:
				sqlConnection.Close()
				sqlConnection.Dispose()
		if exception is not None:
			raise exception

	def ExecuteWithScalar(self, query:str, parameters:list[any] = None):
		returnValue = None
		exception = None
		sqlConnection = None
		sqlCommand = None
		try:
			sqlConnection = SqlClient.SqlConnection(self.ConnectionString)
			sqlConnection.Open()
			sqlCommand = SqlClient.SqlCommand(query, sqlConnection)
			sqlCommand.CommandType = CommandType.Text
			sqlCommand.CommandTimeout = 0
			if (parameters is not None):
				_ParametersHelpers.AddParameters(sqlCommand, parameters)
			returnValue = sqlCommand.ExecuteScalar()
		except Exception as e:
			exception = e
		finally:
			if sqlCommand:
				sqlCommand.Dispose()
			if sqlConnection:
				sqlConnection.Close()
				sqlConnection.Dispose()
		if returnValue is not None:
			return returnValue
		elif exception is not None:
			raise exception

	def TruncateTable(self, schema:str, table:str):
		exception = None
		sqlConnection = None
		sqlCommand = None
		try:
			sqlConnection = SqlClient.SqlConnection(self.ConnectionString)
			sqlConnection.Open()
			sqlCommand = SqlClient.SqlCommand(f"TRUNCATE TABLE [{schema}].[{table}]", sqlConnection)
			sqlCommand.CommandType = CommandType.Text
			sqlCommand.CommandTimeout = 0
			sqlCommand.ExecuteNonQuery()
		except Exception as e:
			exception = e
		finally:
			if sqlCommand:
				sqlCommand.Dispose()
			if sqlConnection:
				sqlConnection.Close()
				sqlConnection.Dispose()
		if exception is not None:
			raise exception

	def GetRowCount(self, schema:str, table:str) -> int:
		returnValue:int = None
		exception = None
		sqlConnection = None
		sqlCommand = None
		try:
			sqlConnection = SqlClient.SqlConnection(self.ConnectionString)
			sqlConnection.Open()
			sqlCommand = SqlClient.SqlCommand(f"SELECT COUNT(*) AS [Count] FROM [{schema}].[{table}]", sqlConnection)
			sqlCommand.CommandType = CommandType.Text
			sqlCommand.CommandTimeout = 0
			result = sqlCommand.ExecuteScalar()
			if (isinstance(result, int)):
				returnValue = int(result)
		except Exception as e:
			exception = e
		finally:
			if sqlCommand:
				sqlCommand.Dispose()
			if sqlConnection:
				sqlConnection.Close()
				sqlConnection.Dispose()
		if exception is not None:
			raise exception
		if returnValue is not None:
			return returnValue

class SQLServerBulkData(IBulkData):
	ConnectionString:str = None

	def __init__(self, connectionString:str):
		self.ConnectionString = connectionString

	def GetDataTable(self, schema:str = None, tableOrView:str = None, query:str = None, options:dict = None):
		returnValue = DataTable()
		exception = None
		sqlConnection = None
		sqlCommand = None
		commandText:str = None
		if (schema is not None and tableOrView is not None):
			commandText = f"SELECT * FROM [{schema}].[{tableOrView}]"
		elif (query is not None):
			commandText = query
		if (commandText is None):
			raise ValueError("Either both Schema and Table/View must be provided, or Query must be provided.")
		try:
			sqlConnection = SqlClient.SqlConnection(self.ConnectionString)
			sqlConnection.Open()
			sqlCommand = SqlClient.SqlCommand(commandText, sqlConnection)
			sqlCommand.CommandType = CommandType.Text
			sqlCommand.CommandTimeout = 0
			returnValue.Load(sqlCommand.ExecuteReader())
		except Exception as e:
			exception = e
		finally:
			if sqlCommand:
				sqlCommand.Dispose()
			if sqlConnection:
				sqlConnection.Close()
				sqlConnection.Dispose()
		if returnValue is not None:
			return returnValue
		elif exception is not None:
			raise exception

	def WriteDataTable(self, schema:str, table:str, dataTable:DataTable, options:dict = { "BatchSize": 1000 }):
		exception = None
		sqlConnection = None
		sqlBulkCopy = None
		try:
			batchSize:int = 1000
			if ("BatchSize" in options):
				batchSize = int(options["BatchSize"])
			sqlConnection = SqlClient.SqlConnection(self.ConnectionString)
			sqlConnection.Open()
			sqlBulkCopy = SqlClient.SqlBulkCopy(sqlConnection)
			sqlBulkCopy.BulkCopyTimeout = 0
			sqlBulkCopy.EnableStreaming = True
			sqlBulkCopy.BatchSize = batchSize
			sqlBulkCopy.DestinationTableName = f"[{schema}].[{table}]"
			for column in dataTable.Columns:
				sqlBulkCopy.ColumnMappings.Add(SqlClient.SqlBulkCopyColumnMapping(column.ColumnName, f"[{column.ColumnName}]"))
			sqlBulkCopy.WriteToServer(dataTable)
		except Exception as e:
			exception = e
		finally:
			if sqlBulkCopy:
				sqlBulkCopy.Close()
			if sqlConnection:
				sqlConnection.Close()
				sqlConnection.Dispose()
		if exception is not None:
			raise exception

class SQLServerClearData(IClearData):
	ConnectionString:str = None

	def __init__(self, connectionString:str):
		self.ConnectionString = connectionString
		self.StoredProcedures = SQLServerStoredProcedures(connectionString)
		self.Queries = Queries(connectionString)
		self.BulkData = BulkData(connectionString)

	@staticmethod
	def ParseParameters(parameters):
		for parameter in parameters:
			print(f"{ParameterDirection.Input} [{parameter.SqlDbType}] -- Name: {parameter.Name} Direction: {parameter.Direction} Value: {parameter.Value}")

	def __GetTestConnectionQuery__(self) -> str:
		return """
			SET NOCOUNT ON
			DECLARE @TimeZone VARCHAR(50)
			EXEC [master].[dbo].[xp_regread]
				N'HKEY_LOCAL_MACHINE','SYSTEM\\CurrentControlSet\\Control\\TimeZoneInformation',
				N'TimeZoneKeyName',
				@TimeZone OUTPUT
			SET NOCOUNT OFF
			SELECT
				CONVERT
				(
					[nvarchar](MAX),
					(
						SELECT
							JSON_QUERY((
								SELECT
									HOST_NAME() AS [HostName],
									APP_NAME() AS [ApplicationName]
									FOR JSON PATH, WITHOUT_ARRAY_WRAPPER, INCLUDE_NULL_VALUES
							)) AS [Client],
							JSON_QUERY((
								SELECT
									@@VERSION AS [Version],
									SERVERPROPERTY(N'ProductVersion') AS [ProductVersion],
									SERVERPROPERTY(N'ServerName') AS [Name],
									SERVERPROPERTY(N'InstanceName') AS [InstanceName],
									SUSER_SNAME() AS [LoginName],
									SERVERPROPERTY(N'Collation') AS [Collation],
									SERVERPROPERTY(N'Edition') AS [Edition],
									SERVERPROPERTY(N'InstanceDefaultBackupPath') AS [InstanceDefaultBackupPath],
									SERVERPROPERTY(N'InstanceDefaultDataPath') AS [InstanceDefaultDataPath],
									SERVERPROPERTY(N'InstanceDefaultLogPath') AS [InstanceDefaultLogPath],
									CASE TRY_CAST(SERVERPROPERTY(N'FilestreamConfiguredLevel') AS [int])
										WHEN 0 THEN N'Disabled'
										WHEN 1 THEN N'Enabled for Transact-SQL access'
										WHEN 2 THEN N'Enabled for Transact-SQL and local Win32 streaming access'
										WHEN 3 THEN N'Enabled for Transact-SQL and both local and remote Win32 streaming access'
									END AS [FilestreamConfiguredLevel],
									CASE TRY_CAST(SERVERPROPERTY(N'FilestreamEffectiveLevel') AS [int])
										WHEN 0 THEN N'Disabled'
										WHEN 1 THEN N'Enabled for Transact-SQL access'
										WHEN 2 THEN N'Enabled for Transact-SQL and local Win32 streaming access'
										WHEN 3 THEN N'Enabled for Transact-SQL and both local and remote Win32 streaming access'
									END AS [FilestreamEffectiveLevel],
									SERVERPROPERTY(N'FilestreamShareName') AS [FilestreamShareName],
									DEFAULT_DOMAIN() AS [DefaultDomain],
									@TimeZone AS [ServerTimeZone]	
									FOR JSON PATH, WITHOUT_ARRAY_WRAPPER, INCLUDE_NULL_VALUES
							)) AS [Server],
							JSON_QUERY((
								SELECT
									DB_NAME() AS [Name],
									USER_NAME() AS [UserName],
									DATABASEPROPERTYEX(DB_NAME(), N'Collation') AS [Collation],
									DATABASEPROPERTYEX(DB_NAME(), N'Edition') AS [Edition],
									DATABASEPROPERTYEX(DB_NAME(), N'Recovery') AS [Recovery],
									DATABASEPROPERTYEX(DB_NAME(), N'Status') AS [Status],
									DATABASEPROPERTYEX(DB_NAME(), N'UserAccess') AS [UserAccess],
									DATABASEPROPERTYEX(DB_NAME(), N'Version') AS [Version]
									FOR JSON PATH, WITHOUT_ARRAY_WRAPPER, INCLUDE_NULL_VALUES
							)) AS [Databse]
							FOR JSON PATH, WITHOUT_ARRAY_WRAPPER, INCLUDE_NULL_VALUES
					),
					0
				) AS [JSON]
		"""

	def TestConnection(self) -> dict:
		returnValue:dict = None
		commandText:str = self.__GetTestConnectionQuery__()
		exception = None
		sqlConnection = None
		sqlCommand = None
		try:
			sqlConnection = SqlClient.SqlConnection(self.ConnectionString)
			sqlConnection.Open()
			sqlCommand = SqlClient.SqlCommand(commandText, sqlConnection)
			sqlCommand.CommandType = CommandType.Text
			sqlCommand.CommandTimeout = 0
			result = sqlCommand.ExecuteScalar()
			if (result):
				returnValue = json.loads(result)
		except Exception as e:
			exception = e
		finally:
			if sqlCommand:
				sqlCommand.Dispose()
			if sqlConnection:
				sqlConnection.Close()
				sqlConnection.Dispose()
		if exception is not None:
			raise exception
		return returnValue

__all__ = ["SQLServerParameter", "SQLServerStoredProcedures", "SQLServerQueries", "SQLServerBulkData", "SQLServerClearData"]